#include "OgreVisibilityOctreeSceneManager.h"
#include "OgreVisibilityOptionsManager.h"
#include <OgreMath.h>
#include <OgreIteratorWrappers.h>
#include <OgreRenderSystem.h>
#include <OgreCamera.h>
#include <OgreLogManager.h>

namespace Ogre {

//-----------------------------------------------------------------------
VisibilityOctreeSceneManager::VisibilityOctreeSceneManager(
	GtpVisibility::VisibilityManager *visManager)
: 
mVisibilityManager(visManager), 
mUseDepthPass(false),
mShowVisualization(false),
mRenderNodesForViz(false),
mVisualizeCulledNodes(false),
mSkipTransparents(false),
mDelayRenderTransparents(true)
{
	mHierarchyInterface = 
		new OctreeHierarchyInterface(this, mDestRenderSystem);
		
	//mDisplayNodes = true;
	//mShowBoundingBoxes = true;
	//mShowBoxes = true;

	// TODO: find reasonable value for max depth
	mMaxDepth = 50;

	CreateDepthPass();
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::CreateDepthPass()
{
	MaterialPtr depthMat = MaterialManager::getSingleton().getByName("Visibility/DepthPass");

	if (depthMat.isNull())
    {
		// Init
		depthMat = MaterialManager::getSingleton().create(
                "Visibility/DepthPass",
                ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);

		depthMat->createTechnique()->createPass();
        mDepthPass = depthMat->getTechnique(0)->getPass(0);
		mDepthPass->setColourWriteEnabled(false);
		mDepthPass->setDepthWriteEnabled(true);
		mDepthPass->setLightingEnabled(false);
	}
	else
	{
		mDepthPass = depthMat->getTechnique(0)->getPass(0);
	}
}
//-----------------------------------------------------------------------
VisibilityOctreeSceneManager::~VisibilityOctreeSceneManager()
{
	delete mHierarchyInterface;
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::_renderVisibleObjects()
{
		if (!mShowVisualization)
	{
		// two cameras (one for culling, one for rendering)
		mHierarchyInterface->InitFrame(mOctree, mCameraInProgress, 
								mCullCamera ? getCamera("CullCamera") : NULL, mDelayRenderTransparents);

		// call initframe to reset culling manager stats
		mVisibilityManager->GetCullingManager()->InitFrame(mVisualizeCulledNodes);
	}
	
	// standard terrain scenemanager rendering without hierarchical culling
	if (!mUseDepthPass || mShowVisualization)
	{	
		OctreeSceneManager::_renderVisibleObjects();
	
		return;
	} 
	
	//-- hierarchical culling
	// the objects of different layers (e.g., background, scene, 
	// overlay) must be identified and rendered one after another

	mSkipTransparents = false;

	//-- render background
	clearSpecialCaseRenderQueues();
	addSpecialCaseRenderQueue(RENDER_QUEUE_BACKGROUND);
	addSpecialCaseRenderQueue(RENDER_QUEUE_SKIES_EARLY);

	setSpecialCaseRenderQueueMode(SceneManager::SCRQM_INCLUDE);
	SceneManager::_renderVisibleObjects();

#ifdef GTP_VISIBILITY_MODIFIED_OGRE
	_deleteRenderedQueueGroups(false);
#endif

	//-- render visible objects (i.e., all but overlay)
	clearSpecialCaseRenderQueues();
	addSpecialCaseRenderQueue(RENDER_QUEUE_OVERLAY);
	setSpecialCaseRenderQueueMode(SceneManager::SCRQM_EXCLUDE);
	
	// render only solids in hierarchical algorithm. 
	// Transparents need sorting, thus we render them afterwards
	mSkipTransparents = mDelayRenderTransparents;

	/** 
	* the hierarchical culling algorithm
	**/
	
	mVisibilityManager->ApplyVisibilityCulling();
	
	//-- now we can savely render all remaining objects, e.g., transparents, overlay
	mSkipTransparents = false;

	clearSpecialCaseRenderQueues();
	SceneManager::_renderVisibleObjects();

	//WriteLog(); // write out stats
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::_findVisibleObjects(Camera* cam, bool onlyShadowCasters)
{
	// only shadow casters will be rendered in shadow texture pass
	mHierarchyInterface->SetOnlyShadowCasters(onlyShadowCasters);

	// does nothing if hierarchical culling is used =>
	// we interleave identification and rendering of objects 
	// in _renderVisibibleObjects
	if (!mUseDepthPass)
	{
		OctreeSceneManager::_findVisibleObjects(cam, onlyShadowCasters);
		return;	
	}

	//-- show visibile scene nodes and octree bounding boxes from last frame
	if (mShowVisualization)
    {
		// add player camera for visualization purpose
		try {
			Camera *c;
			if ((c = getCamera("PlayerCam")) != NULL)
			{
				getRenderQueue()->addRenderable(c);
			}   
        }
        catch(...)
        {
            // ignore
        }
		
		if (mRenderNodesForViz || mRenderNodesContentForViz)
		{
		/*	// change node material so it is better suited for visualization
			MaterialPtr nodeMat = MaterialManager::getSingleton().getByName("Core/NodeMaterial");
			nodeMat->setAmbient(1, 1, 0);
			nodeMat->setLightingEnabled(true);
			nodeMat->getTechnique(0)->getPass(0)->removeAllTextureUnitStates();*/

			for (NodeList::iterator it = mVisible.begin(); it != mVisible.end(); ++it)
			{
				if (mRenderNodesForViz)
				{
					getRenderQueue()->addRenderable(*it);
				}
				if (mRenderNodesContentForViz) 
				{
					(*it)->_addToRenderQueue(cam, getRenderQueue(), false);
				}
			}
		}
		for (BoxList::iterator it = mBoxes.begin(); it != mBoxes.end(); ++it)
		{
			getRenderQueue()->addRenderable(*it);
		}
	}
	
	mVisible.clear();
    mBoxes.clear();
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::_updateSceneGraph(Camera* cam)
{
	mVisibilityManager->GetCullingManager()->SetHierarchyInterface(mHierarchyInterface);
	mHierarchyInterface->SetRenderSystem(mDestRenderSystem);

#ifdef GTP_VISIBILITY_MODIFIED_OGRE
    mHierarchyInterface->SetNumOctreeNodes(mNumOctreeNodes);
#endif
	OctreeSceneManager::_updateSceneGraph(cam);
}
//-----------------------------------------------------------------------
bool VisibilityOctreeSceneManager::setOption(const String & key, const void * val)
{
	if (key == "UseDepthPass")
	{
		mUseDepthPass = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "ShowVisualization")
	{
		mShowVisualization = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "RenderNodesForViz")
	{
		mRenderNodesForViz = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "SkyBoxEnabled")
	{
		mSkyBoxEnabled = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "SkyPlaneEnabled")
	{
		mSkyPlaneEnabled = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "SkyDomeEnabled")
	{
		mSkyDomeEnabled = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "VisualizeCulledNodes")
	{
		mVisualizeCulledNodes = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "DelayRenderTransparents")
	{
		mDelayRenderTransparents = (*static_cast<const bool *>(val));
		return true;
	}
	return VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface).
		setOption(key, val) || OctreeSceneManager::setOption(key, val);
}
//-----------------------------------------------------------------------
bool VisibilityOctreeSceneManager::getOption(const String & key, void *val)
{
	if (key == "NumHierarchyNodes")
	{
		* static_cast<unsigned int *>(val) = (unsigned int)mNumOctreeNodes;
		return true;
	}

	return VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface).
		getOption(key, val) && OctreeSceneManager::getOption(key, val);
}
//-----------------------------------------------------------------------
bool VisibilityOctreeSceneManager::getOptionValues(const String & key, StringVector  &refValueList)
{
	return OctreeSceneManager::getOptionValues( key, refValueList );
}
//-----------------------------------------------------------------------
bool VisibilityOctreeSceneManager::getOptionKeys(StringVector & refKeys)
{
	return  VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface).
		getOptionKeys (refKeys) || OctreeSceneManager::getOptionKeys(refKeys);
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::setVisibilityManager(GtpVisibility::VisibilityManager *visManager)
{
    mVisibilityManager = visManager;
}
//-----------------------------------------------------------------------
GtpVisibility::VisibilityManager *VisibilityOctreeSceneManager::getVisibilityManager()
{
	return mVisibilityManager;
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::renderObjects(const RenderPriorityGroup::TransparentRenderablePassList& objs, 
            bool doLightIteration, const LightList* manualLightList)
{
	if (!mSkipTransparents)
	{
		OctreeSceneManager::renderObjects(objs, doLightIteration, manualLightList);
	}
}
//-----------------------------------------------------------------------
void VisibilityOctreeSceneManager::WriteLog()
{
	std::stringstream d;

	d << "Algorithm: " << mVisibilityManager->GetCullingManagerType() << ", "
	  << "Hierarchy nodes: " << mNumOctreeNodes << ", " << "Traversed nodes: "
	  << mHierarchyInterface->GetNumTraversedNodes() << ", "
	  << "Rendered nodes: " << mHierarchyInterface->GetNumRenderedNodes() << ", "
	  << "Query culled nodes: " << mVisibilityManager->GetCullingManager()->GetNumQueryCulledNodes() << ", "
	  << "Frustum culled nodes: " << mVisibilityManager->GetCullingManager()->GetNumFrustumCulledNodes() << ", "
      << "Queries issued: " << mVisibilityManager->GetCullingManager()->GetNumQueriesIssued() << "\n";

	LogManager::getSingleton().logMessage(d.str());
}
}   // namespace Ogre